import asyncio
import json
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from blackgeorge.adapters.base import ModelResponse
from blackgeorge.core.event import Event
from blackgeorge.core.event_types import EventType
from blackgeorge.core.job import Job
from blackgeorge.core.message import Message
from blackgeorge.core.pending_action import PendingAction
from blackgeorge.core.report import Report
from blackgeorge.core.tool_call import ToolCall
from blackgeorge.core.types import RunStatus
from blackgeorge.store.state import RunState
from blackgeorge.tools.base import Tool, ToolResult
from blackgeorge.tools.execution import aexecute_tool
from blackgeorge.utils import new_id
from blackgeorge.worker_context import (
    SUMMARY_ATTEMPT_LIMIT,
    context_error_message,
)
from blackgeorge.worker_messages import (
    emit_assistant_message,
    replace_tool_call,
    structured_content,
    tool_call_with_result,
    tool_message,
)
from blackgeorge.worker_tools import (
    pending_options,
    tool_action_type,
    tool_prompt,
)

if TYPE_CHECKING:
    from blackgeorge.config import RunConfig

EventEmitter = Callable[[str, str, dict[str, Any]], None]


def _build_report(
    run_id: str,
    status: RunStatus,
    content: str | None = None,
    reasoning_content: str | None = None,
    data: Any | None = None,
    messages: list[Message] | None = None,
    tool_calls: list[ToolCall] | None = None,
    metrics: dict[str, Any] | None = None,
    events: list[Event] | None = None,
    pending_action: PendingAction | None = None,
    errors: list[str] | None = None,
) -> Report:
    return Report(
        run_id=run_id,
        status=status,
        content=content,
        reasoning_content=reasoning_content,
        data=data,
        messages=list(messages) if messages else [],
        tool_calls=list(tool_calls) if tool_calls else [],
        metrics=metrics or {},
        events=list(events) if events else [],
        pending_action=pending_action,
        errors=list(errors) if errors else [],
    )


def _build_state(
    run_id: str,
    status: RunStatus,
    runner_name: str,
    job: Job,
    messages: list[Message],
    tool_calls: list[ToolCall],
    pending_action: PendingAction | None,
    metrics: dict[str, Any],
    iteration: int,
    payload: dict[str, Any] | None = None,
) -> RunState:
    return RunState(
        run_id=run_id,
        status=status,
        runner_type="worker",
        runner_name=runner_name,
        job=job,
        messages=messages,
        tool_calls=tool_calls,
        pending_action=pending_action,
        metrics=metrics,
        iteration=iteration,
        payload=payload or {},
    )


def _report_error(
    run_id: str, messages: list[Message], errors: list[str], events: list[Event]
) -> Report:
    return Report(
        run_id=run_id,
        status="failed",
        content=None,
        reasoning_content=None,
        data=None,
        messages=messages,
        tool_calls=[],
        metrics={},
        events=events,
        pending_action=None,
        errors=errors,
    )


def _ensure_not_running_loop(action: str, async_action: str) -> None:
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return
    raise RuntimeError(
        f"{action} cannot be called from a running event loop. Use {async_action} instead."
    )


def _fail_report(
    *,
    config: "RunConfig",
    worker_name: str,
    message: str,
    messages: list[Message],
    tool_calls: list[ToolCall],
    metrics: dict[str, Any],
    errors: list[str],
) -> Report:
    errors.append(message)
    config.emit(EventType.WORKER_FAILED, worker_name, {"error": message})
    return _build_report(
        config.run_id,
        "failed",
        None,
        None,
        None,
        messages,
        tool_calls,
        metrics,
        config.events,
        None,
        errors,
    )


async def _acontext_retry(
    *,
    config: "RunConfig",
    worker_name: str,
    messages: list[Message],
    tool_calls: list[ToolCall],
    metrics: dict[str, Any],
    errors: list[str],
    model_registered: bool,
    context_summaries: int,
    apply_summary: Callable[[], Awaitable[bool]],
) -> "ContextDecision":
    if not config.respect_context_window:
        return ContextDecision(
            False,
            _fail_report(
                config=config,
                worker_name=worker_name,
                message=context_error_message(model_registered, False),
                messages=messages,
                tool_calls=tool_calls,
                metrics=metrics,
                errors=errors,
            ),
        )
    if context_summaries >= SUMMARY_ATTEMPT_LIMIT:
        return ContextDecision(
            False,
            _fail_report(
                config=config,
                worker_name=worker_name,
                message=context_error_message(model_registered, True),
                messages=messages,
                tool_calls=tool_calls,
                metrics=metrics,
                errors=errors,
            ),
        )
    if not await apply_summary():
        return ContextDecision(
            False,
            _fail_report(
                config=config,
                worker_name=worker_name,
                message=context_error_message(model_registered, True),
                messages=messages,
                tool_calls=tool_calls,
                metrics=metrics,
                errors=errors,
            ),
        )
    return ContextDecision(True, None)


def _should_stream(stream: bool, _tools: list[Tool], response_schema: Any | None) -> bool:
    return stream and response_schema is None


def _tool_result_preview(result: ToolResult, limit: int) -> tuple[str | None, bool]:
    if result.content is not None:
        text = result.content
    elif result.data is not None:
        try:
            text = json.dumps(result.data, ensure_ascii=True)
        except (TypeError, ValueError):
            text = str(result.data)
    elif result.error is not None:
        text = result.error
    else:
        return None, False
    if len(text) > limit:
        return f"{text[:limit]}...", True
    return text, False


def _tool_event_payload(call: ToolCall, result: ToolResult, limit: int = 200) -> dict[str, Any]:
    payload: dict[str, Any] = {"tool_call_id": call.id}
    preview, truncated = _tool_result_preview(result, limit)
    if preview is not None:
        payload["result_preview"], payload["result_truncated"] = preview, truncated
    if result.timed_out:
        payload["timed_out"] = True
    if result.cancelled:
        payload["cancelled"] = True
    return payload


def _record_usage(metrics: dict[str, Any], response: object) -> None:
    if hasattr(response, "usage"):
        metrics["usage"] = response.usage
    else:
        metrics["usage"] = {}


def _finalize_structured_response(
    *,
    config: "RunConfig",
    data: Any,
    messages: list[Message],
    tool_calls: list[ToolCall],
    metrics: dict[str, Any],
    errors: list[str],
    worker_name: str,
) -> Report:
    content = structured_content(data)
    assistant_message = Message(role="assistant", content=content)
    messages.append(assistant_message)
    emit_assistant_message(config.emit, worker_name, assistant_message)
    config.emit(EventType.WORKER_COMPLETED, worker_name, {})
    return _build_report(
        config.run_id,
        "completed",
        content,
        None,
        data,
        messages,
        tool_calls,
        metrics,
        config.events,
        None,
        errors,
    )


def _finalize_plain_response(
    *,
    config: "RunConfig",
    response: ModelResponse,
    messages: list[Message],
    tool_calls: list[ToolCall],
    metrics: dict[str, Any],
    errors: list[str],
    worker_name: str,
) -> Report:
    assistant_message = Message(
        role="assistant",
        content=response.content or "",
        reasoning_content=response.reasoning_content,
        thinking_blocks=response.thinking_blocks,
    )
    messages.append(assistant_message)
    emit_assistant_message(config.emit, worker_name, assistant_message)
    config.emit(EventType.WORKER_COMPLETED, worker_name, {})
    return _build_report(
        config.run_id,
        "completed",
        response.content,
        response.reasoning_content,
        None,
        messages,
        tool_calls,
        metrics,
        config.events,
        None,
        errors,
    )


@dataclass(frozen=True)
class ToolPlan:
    ordered_calls: list[ToolCall]
    executable_calls: list[tuple[ToolCall, Tool]]
    immediate_results: dict[str, ToolResult]
    pending: PendingAction | None
    max_tool_calls_exceeded: bool


@dataclass(frozen=True)
class ContextDecision:
    retry: bool
    report: Report | None


def _plan_tool_calls(
    *,
    response: ModelResponse,
    allowed_tools: dict[str, Tool],
    tool_calls: list[ToolCall],
    max_tool_calls: int,
) -> ToolPlan:
    ordered_calls, executable_calls, immediate_results, pending, max_tool_calls_exceeded = (
        [],
        [],
        {},
        None,
        False,
    )
    for call in response.tool_calls:
        if len(tool_calls) >= max_tool_calls:
            max_tool_calls_exceeded = True
            break
        tool_calls.append(call)
        if call.error:
            ordered_calls.append(call)
            immediate_results[call.id] = ToolResult(error=call.error)
            continue
        tool = allowed_tools.get(call.name)
        if tool is None:
            ordered_calls.append(call)
            immediate_results[call.id] = ToolResult(error=f"Tool not found: {call.name}")
            continue
        action_type = tool_action_type(tool)
        if action_type:
            metadata = {"tool": tool.name}
            if tool.input_key:
                metadata["input_key"] = tool.input_key
            pending = PendingAction(
                action_id=new_id(),
                type=action_type,
                tool_call=call,
                prompt=tool_prompt(tool, action_type, call),
                options=pending_options(action_type),
                metadata=metadata,
            )
            break
        ordered_calls.append(call)
        executable_calls.append((call, tool))
    return ToolPlan(
        ordered_calls, executable_calls, immediate_results, pending, max_tool_calls_exceeded
    )


async def _execute_tool_calls_async(
    config: "RunConfig",
    ordered_calls: list[ToolCall],
    executable_calls: list[tuple[ToolCall, Tool]],
    immediate_results: dict[str, ToolResult],
    messages: list[Message],
    tool_calls: list[ToolCall],
) -> None:
    results = dict(immediate_results)
    if executable_calls:
        for call, tool in executable_calls:
            config.emit(EventType.TOOL_STARTED, tool.name, {"tool_call_id": call.id})
        if len(executable_calls) == 1:
            call, tool = executable_calls[0]
            results[call.id] = await aexecute_tool(tool, call)
        else:
            tasks = [aexecute_tool(tool, call) for call, tool in executable_calls]
            tool_results = await asyncio.gather(*tasks)
            for (call, _), result in zip(executable_calls, tool_results, strict=True):
                results[call.id] = result
    for call in ordered_calls:
        result = results.get(call.id, ToolResult(error="Tool execution failed"))
        if result.error:
            config.emit(
                EventType.TOOL_FAILED, call.name, {"tool_call_id": call.id, "error": result.error}
            )
        else:
            config.emit(EventType.TOOL_COMPLETED, call.name, _tool_event_payload(call, result))
        messages.append(tool_message(result, call))
        replace_tool_call(tool_calls, tool_call_with_result(call, result))



from blackgeorge.workflow.context import WorkflowContext
from blackgeorge.workflow.flow import Flow
from blackgeorge.workflow.nodes import (
    AsyncCondition,
    AsyncLoop,
    Condition,
    Loop,
    Parallel,
    Router,
    Step,
)
from blackgeorge.workflow.result import StepResult

__all__ = [
    "AsyncCondition",
    "AsyncLoop",
    "Condition",
    "Flow",
    "Loop",
    "Parallel",
    "Router",
    "Step",
    "StepResult",
    "WorkflowContext",
]



import json
from collections.abc import Iterable
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field

from blackgeorge.core.job import Job
from blackgeorge.tools.base import Tool, ToolResult
from blackgeorge.tools.registry import Toolbelt
from blackgeorge.utils import new_id
from blackgeorge.worker import Worker

if TYPE_CHECKING:
    from blackgeorge.desk import Desk


class SpawnSubworkerInput(BaseModel):
    name: str = Field(..., description="Name of the subworker.")
    instructions: str = Field(..., description="Instructions for the subworker.")
    task: str = Field(..., description="Task input for the subworker.")
    tools: list[str] = Field(
        default_factory=list,
        description="Names of tools the subworker is allowed to use.",
    )
    model: str | None = Field(
        default=None,
        description="Model override for the subworker. Defaults to configured default.",
    )


def _join_errors(errors: list[str], fallback: str) -> str:
    if not errors:
        return fallback
    text = "; ".join(error for error in errors if error)
    return text or fallback


def _child_desk(
    desk: "Desk",
    model: str,
    *,
    structured_output_retries: int,
    max_iterations: int,
    max_tool_calls: int,
) -> "Desk":
    from blackgeorge.desk import Desk

    return Desk(
        model=model,
        temperature=desk.temperature,
        max_tokens=desk.max_tokens,
        stream=False,
        structured_stream_mode=desk.structured_stream_mode,
        structured_output_retries=structured_output_retries,
        max_iterations=max_iterations,
        max_tool_calls=max_tool_calls,
        respect_context_window=desk.respect_context_window,
        event_bus=desk.event_bus,
        run_store=desk.run_store,
        memory_store=desk.memory_store,
        adapter=desk.adapter,
        storage_dir=desk.storage_dir,
    )


def _mark_child_run_failed(
    desk: "Desk",
    run_id: str,
    job: Job,
    error: str,
) -> None:
    record = desk.run_store.get_run(run_id)
    if record is None:
        desk.run_store.create_run(run_id, job.model_dump(mode="json"))
        record = desk.run_store.get_run(run_id)
    if record is not None and record.status != "running":
        return
    events = desk.run_store.get_events(run_id)
    desk._emit(
        events,
        run_id,
        "run.failed",
        "desk",
        {"errors": [error]},
    )
    desk.run_store.update_run(
        run_id,
        "failed",
        None,
        {"error": error},
        None,
    )


def create_subworker_tool(
    *,
    desk: "Desk",
    available_tools: list[Tool] | None = None,
    default_model: str | None = None,
    allowed_models: Iterable[str] | None = None,
    max_subworkers: int = 5,
    max_tools_per_subworker: int = 8,
    max_iterations: int = 10,
    max_tool_calls: int = 10,
    structured_output_retries: int = 1,
    tool_name: str = "spawn_subworker",
) -> Tool:
    if max_subworkers < 1:
        raise ValueError("max_subworkers must be >= 1")
    if max_tools_per_subworker < 0:
        raise ValueError("max_tools_per_subworker must be >= 0")
    if max_iterations < 1:
        raise ValueError("max_iterations must be >= 1")
    if max_tool_calls < 1:
        raise ValueError("max_tool_calls must be >= 1")

    allowed_model_set = set(allowed_models) if allowed_models is not None else None
    toolbelt = Toolbelt(available_tools or [])
    spawn_count = 0

    async def spawn_subworker(
        name: str,
        instructions: str,
        task: str,
        tools: list[str] | None = None,
        model: str | None = None,
    ) -> ToolResult:
        nonlocal spawn_count

        if spawn_count >= max_subworkers:
            return ToolResult(
                error=f"Subworker budget exceeded: max_subworkers={max_subworkers}",
            )

        resolved_model = model or default_model or desk.model
        if not resolved_model:
            return ToolResult(error="Subworker model is not configured")

        if allowed_model_set is not None and resolved_model not in allowed_model_set:
            return ToolResult(error=f"Model '{resolved_model}' is not allowed for subworkers")

        requested_tools = tools or []
        if len(requested_tools) > max_tools_per_subworker:
            return ToolResult(
                error=(
                    "Too many tools requested for subworker: "
                    f"{len(requested_tools)} > {max_tools_per_subworker}"
                )
            )

        worker_tools: list[Tool] = []
        missing_tools: list[str] = []
        for tool_name_value in requested_tools:
            resolved_tool = toolbelt.resolve(tool_name_value)
            if resolved_tool is None:
                missing_tools.append(tool_name_value)
            else:
                worker_tools.append(resolved_tool)
        if missing_tools:
            missing = ", ".join(missing_tools)
            return ToolResult(error=f"Requested tools are not available: {missing}")

        subworker = Worker(
            name=name,
            instructions=instructions,
            model=resolved_model,
            tools=worker_tools,
        )
        child_runner = _child_desk(
            desk,
            resolved_model,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
        )
        child_job = Job(input=task)
        child_run_id = new_id()

        try:
            report = await child_runner.arun(
                subworker,
                child_job,
                stream=False,
                run_id=child_run_id,
            )
        except Exception as exc:
            error_message = str(exc)
            _mark_child_run_failed(
                desk=child_runner,
                run_id=child_run_id,
                job=child_job,
                error=error_message,
            )
            return ToolResult(
                error=f"Subworker execution failed: {error_message}",
                data={
                    "run_id": child_run_id,
                    "status": "failed",
                    "worker": name,
                    "errors": [error_message],
                },
            )

        result_data: dict[str, object] = {
            "run_id": report.run_id,
            "status": report.status,
            "worker": name,
        }
        if report.errors:
            result_data["errors"] = list(report.errors)
        if report.pending_action is not None:
            result_data["pending_action_type"] = report.pending_action.type

        if report.status == "paused":
            if report.pending_action is None:
                pending_type = "unknown"
            else:
                pending_type = report.pending_action.type
            return ToolResult(
                error=f"Subworker paused and requires {pending_type} follow-up",
                data=result_data,
            )

        if report.status == "failed":
            return ToolResult(
                error=_join_errors(report.errors, "Subworker failed"),
                data=result_data,
            )

        spawn_count += 1
        content = report.content
        if content is None and report.data is not None:
            content = json.dumps(report.data, ensure_ascii=True, default=str)
        if content is None:
            content = ""

        return ToolResult(content=content, data=result_data)

    return Tool(
        name=tool_name,
        description="Spawn a bounded autonomous subworker for delegated work.",
        schema=SpawnSubworkerInput.model_json_schema(),
        callable=spawn_subworker,
        input_model=SpawnSubworkerInput,
    )


def create_swarm_tool(
    *,
    desk: "Desk",
    available_tools: list[Tool] | None = None,
    default_model: str | None = None,
    allowed_models: Iterable[str] | None = None,
    max_subworkers: int = 5,
    max_tools_per_subworker: int = 8,
    max_iterations: int = 10,
    max_tool_calls: int = 10,
    structured_output_retries: int = 1,
) -> Tool:
    return create_subworker_tool(
        desk=desk,
        available_tools=available_tools,
        default_model=default_model,
        allowed_models=allowed_models,
        max_subworkers=max_subworkers,
        max_tools_per_subworker=max_tools_per_subworker,
        max_iterations=max_iterations,
        max_tool_calls=max_tool_calls,
        structured_output_retries=structured_output_retries,
        tool_name="spawn_agent",
    )



from typing import Any

from blackgeorge.config import RunConfig
from blackgeorge.core.job import Job
from blackgeorge.core.report import Report
from blackgeorge.store.state import RunState
from blackgeorge.tools.base import Tool
from blackgeorge.tools.registry import Toolbelt
from blackgeorge.worker_runner import WorkerRunner


class Worker:
    def __init__(
        self,
        name: str,
        tools: list[Tool] | None = None,
        model: str | None = None,
        instructions: str | None = None,
        memory_scope: str | None = None,
    ) -> None:
        self.name = name
        self.model = model
        self.instructions = instructions
        self.toolbelt = Toolbelt(tools)
        self.memory_scope = memory_scope or f"worker:{name}"

    def tools(self) -> list[Tool]:
        return self.toolbelt.list()

    def _runner(self) -> WorkerRunner:
        return WorkerRunner(self.name, self.toolbelt, self.instructions)

    def run(self, config: RunConfig, job: Job) -> tuple[Report, RunState | None]:
        return self._runner().run(config, job, self.model)

    async def arun(self, config: RunConfig, job: Job) -> tuple[Report, RunState | None]:
        return await self._runner().arun(config, job, self.model)

    def resume(
        self, config: RunConfig, state: RunState, decision_or_input: Any
    ) -> tuple[Report, RunState | None]:
        return self._runner().resume(config, state, decision_or_input, self.model)

    async def aresume(
        self, config: RunConfig, state: RunState, decision_or_input: Any
    ) -> tuple[Report, RunState | None]:
        return await self._runner().aresume(config, state, decision_or_input, self.model)



from blackgeorge.runner.streaming import (
    append_tool_error,
    chunk_tool_call_deltas,
    is_stream_unsupported_error,
    parse_structured_stream_json,
    stream_value,
    streamed_tool_calls,
)

__all__ = [
    "append_tool_error",
    "chunk_tool_call_deltas",
    "is_stream_unsupported_error",
    "parse_structured_stream_json",
    "stream_value",
    "streamed_tool_calls",
]



import asyncio
import queue
import threading
from collections.abc import AsyncIterator, Iterator
from typing import Any, cast

from pydantic import BaseModel, ConfigDict

from blackgeorge.core.event import Event
from blackgeorge.core.job import Job
from blackgeorge.core.message import Message
from blackgeorge.core.report import Report
from blackgeorge.desk import Desk
from blackgeorge.store.session_store import SessionStore
from blackgeorge.store.sqlite_session_store import SQLiteSessionStore
from blackgeorge.utils import new_id
from blackgeorge.worker import Worker


class WorkerSession(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    session_id: str
    worker: Worker
    desk: Desk
    store: SessionStore

    @classmethod
    def start(
        cls,
        *,
        worker: Worker,
        desk: Desk,
        session_id: str | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> "WorkerSession":
        store = SQLiteSessionStore(desk.db_path)
        session_id = session_id or new_id()

        store.create_session(
            session_id=session_id,
            worker_name=worker.name,
            metadata=metadata or {},
        )

        return cls(
            session_id=session_id,
            worker=worker,
            desk=desk,
            store=store,
        )

    @classmethod
    def resume(
        cls,
        *,
        session_id: str,
        worker: Worker,
        desk: Desk,
    ) -> "WorkerSession | None":
        store = SQLiteSessionStore(desk.db_path)

        record = store.get_session(session_id)
        if record is None or record.worker_name != worker.name:
            return None

        return cls(
            session_id=session_id,
            worker=worker,
            desk=desk,
            store=store,
        )

    def run(self, user_input: Any, *, stream: bool = False, **job_kwargs: Any) -> Report:
        job, initial_count = self._build_job(user_input, job_kwargs)
        report = self.desk.run(self.worker, job, stream=stream)
        self._persist_report(report, initial_count)
        return report

    async def arun(self, user_input: Any, *, stream: bool = False, **job_kwargs: Any) -> Report:
        job, initial_count = self._build_job(user_input, job_kwargs)
        report = await self.desk.arun(self.worker, job, stream=stream)
        self._persist_report(report, initial_count)
        return report

    def history(self) -> list[Message]:
        return self.store.get_messages(self.session_id)

    def stream_run(self, user_input: Any, **job_kwargs: Any) -> Iterator[Event]:
        job, initial_count = self._build_job(user_input, job_kwargs)
        run_id = new_id()
        stream_done = object()
        events: queue.Queue[Event | object] = queue.Queue()
        report: Report | None = None
        failure: Exception | None = None
        completed = False

        def handler(event: Event) -> None:
            if event.run_id == run_id:
                events.put(event)

        def run_job() -> None:
            nonlocal report, failure
            try:
                report = self.desk.run(self.worker, job, stream=True, run_id=run_id)
            except Exception as exc:
                failure = exc
            finally:
                events.put(stream_done)

        self.desk.event_bus.subscribe("*", handler)
        thread = threading.Thread(target=run_job, daemon=True)
        thread.start()

        try:
            while True:
                item = events.get()
                if item is stream_done:
                    completed = True
                    break
                yield cast(Event, item)
        finally:
            self.desk.event_bus.unsubscribe("*", handler)
            thread.join()
            if report is not None:
                self._persist_report(report, initial_count)

        if not completed:
            return
        if failure is not None:
            raise failure
        if report is None:
            raise RuntimeError("Run did not produce a report")

    async def astream_run(self, user_input: Any, **job_kwargs: Any) -> AsyncIterator[Event]:
        job, initial_count = self._build_job(user_input, job_kwargs)
        run_id = new_id()
        stream_done = object()
        events: asyncio.Queue[Event | object] = asyncio.Queue()
        completed = False

        def handler(event: Event) -> None:
            if event.run_id == run_id:
                events.put_nowait(event)

        self.desk.event_bus.subscribe("*", handler)
        run_task: asyncio.Task[Report] = asyncio.create_task(
            self.desk.arun(self.worker, job, stream=True, run_id=run_id)
        )
        run_task.add_done_callback(lambda _task: events.put_nowait(stream_done))

        try:
            while True:
                item = await events.get()
                if item is stream_done:
                    completed = True
                    break
                yield cast(Event, item)
        finally:
            self.desk.event_bus.unsubscribe("*", handler)
            if not run_task.done():
                await run_task
            if run_task.done() and not run_task.cancelled() and run_task.exception() is None:
                self._persist_report(run_task.result(), initial_count)

        if not completed:
            return
        if run_task.cancelled():
            raise asyncio.CancelledError
        exception = run_task.exception()
        if exception is not None:
            raise exception

    def close(self) -> None:
        self.store.delete_session(self.session_id)

    def _build_job(self, user_input: Any, job_kwargs: dict[str, Any]) -> tuple[Job, int]:
        messages = self.store.get_messages(self.session_id)
        initial_count = len(messages) if messages else 0
        job = Job(
            input=user_input,
            initial_messages=messages if messages else None,
            **job_kwargs,
        )
        return job, initial_count

    def _persist_report(self, report: Report, initial_count: int) -> None:
        if report.status != "completed":
            return
        new_messages = self._extract_conversation_messages(report)
        self.store.add_messages(self.session_id, new_messages[initial_count:])
        self.store.update_session(self.session_id)

    def _extract_conversation_messages(self, report: Report) -> list[Message]:
        messages: list[Message] = []

        for message in report.messages:
            if message.role in ("user", "assistant") or (
                message.role == "tool" and message.tool_call_id
            ):
                msg_dict = message.model_dump()
                if message.role == "assistant" and not message.tool_calls:
                    msg_dict.pop("reasoning_content", None)
                    msg_dict.pop("thinking_blocks", None)
                messages.append(Message(**msg_dict))

        return messages



from typing import Any

from blackgeorge.core.event import Event
from blackgeorge.core.types import RunStatus
from blackgeorge.store.base import RunRecord, RunStore
from blackgeorge.store.state import RunState
from blackgeorge.utils import utc_now


class InMemoryRunStore(RunStore):
    def __init__(self) -> None:
        self._runs: dict[str, RunRecord] = {}
        self._events: dict[str, list[Event]] = {}

    def create_run(self, run_id: str, input_payload: Any) -> None:
        now = utc_now()
        self._runs[run_id] = RunRecord(
            run_id=run_id,
            status="running",
            input=input_payload,
            output=None,
            output_json=None,
            created_at=now,
            updated_at=now,
            state=None,
        )
        self._events[run_id] = []

    def update_run(
        self,
        run_id: str,
        status: RunStatus,
        output: str | None,
        output_json: Any | None,
        state: RunState | None,
    ) -> None:
        record = self._runs.get(run_id)
        now = utc_now()
        if record is None:
            self._runs[run_id] = RunRecord(
                run_id=run_id,
                status=status,
                input=None,
                output=output,
                output_json=output_json,
                created_at=now,
                updated_at=now,
                state=state,
            )
            return
        self._runs[run_id] = RunRecord(
            run_id=record.run_id,
            status=status,
            input=record.input,
            output=output,
            output_json=output_json,
            created_at=record.created_at,
            updated_at=now,
            state=state,
        )

    def get_run(self, run_id: str) -> RunRecord | None:
        return self._runs.get(run_id)

    def add_event(self, event: Event) -> None:
        self._events.setdefault(event.run_id, []).append(event)

    def get_events(self, run_id: str) -> list[Event]:
        return list(self._events.get(run_id, []))



from datetime import datetime
from typing import Any

from pydantic import BaseModel, ConfigDict, Field


class Event(BaseModel):
    event_id: str
    type: str
    timestamp: datetime
    run_id: str
    source: str
    payload: dict[str, Any] = Field(default_factory=dict)

    model_config = ConfigDict(frozen=True)



from typing import Any

from pydantic import BaseModel, ConfigDict, Field

from blackgeorge.core.job import Job
from blackgeorge.core.message import Message
from blackgeorge.core.pending_action import PendingAction
from blackgeorge.core.tool_call import ToolCall
from blackgeorge.core.types import RunStatus


class RunState(BaseModel):
    run_id: str
    status: RunStatus
    runner_type: str
    runner_name: str
    job: Job
    messages: list[Message] = Field(default_factory=list)
    tool_calls: list[ToolCall] = Field(default_factory=list)
    pending_action: PendingAction | None = None
    metrics: dict[str, Any] = Field(default_factory=dict)
    iteration: int = 0
    payload: dict[str, Any] = Field(default_factory=dict)

    model_config = ConfigDict(arbitrary_types_allowed=True)



from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from pydantic import BaseModel

from blackgeorge.core.tool_call import ToolCall

ToolPreHook = Callable[[ToolCall], Any]
ToolPostHook = Callable[[ToolCall, "ToolResult"], Any]
ProgressCallback = Callable[[str], None]


@dataclass(frozen=True)
class ToolResult:
    content: str | None = None
    data: Any | None = None
    error: str | None = None
    timed_out: bool = False
    cancelled: bool = False
    exception_type: str | None = None


@dataclass(frozen=True)
class Tool:
    name: str
    description: str
    schema: dict[str, Any]
    callable: Callable[..., Any]
    input_model: type[BaseModel]
    requires_confirmation: bool = False
    requires_user_input: bool = False
    requires_handoff: bool = False
    external_execution: bool = False
    pre: tuple[ToolPreHook, ...] = ()
    post: tuple[ToolPostHook, ...] = ()
    confirmation_prompt: str | None = None
    user_input_prompt: str | None = None
    input_key: str | None = None
    timeout: float | None = None
    retries: int = 0
    retry_delay: float = 1.0
    output_type: type[BaseModel] | None = None



import asyncio
from typing import Any


def _apply_litellm_patches() -> None:
    try:
        import litellm
        from litellm.litellm_core_utils import logging_worker
    except ImportError:
        return

    litellm.suppress_debug_info = True

    original_ensure_queue = logging_worker.LoggingWorker._ensure_queue

    def patched_ensure_queue(self: Any) -> None:
        try:
            current_loop = asyncio.get_running_loop()
        except RuntimeError:
            return

        if self._queue is not None and self._bound_loop is not current_loop:
            while not self._queue.empty():
                try:
                    task = self._queue.get_nowait()
                    if task and "coroutine" in task:
                        task["coroutine"].close()
                except asyncio.QueueEmpty:
                    break

        original_ensure_queue(self)

    logging_worker.LoggingWorker._ensure_queue = patched_ensure_queue  # type: ignore[method-assign]


_apply_litellm_patches()



import json
from typing import Any, cast

from pydantic import BaseModel, TypeAdapter

from blackgeorge.core.tool_call import ToolCall
from blackgeorge.utils import new_id


def parse_structured_stream_json(response_schema: Any, content: str) -> Any:
    if isinstance(response_schema, TypeAdapter):
        return response_schema.validate_json(content)
    if isinstance(response_schema, type) and issubclass(response_schema, BaseModel):
        return response_schema.model_validate_json(content)
    return TypeAdapter(response_schema).validate_json(content)


def stream_value(value: Any, key: str, default: Any = None) -> Any:
    if isinstance(value, dict):
        return value.get(key, default)
    return getattr(value, key, default)


def append_tool_error(current: str | None, message: str) -> str:
    return message if current is None else f"{current}; {message}"


def is_stream_unsupported_error(exc: Exception) -> bool:
    if isinstance(exc, NotImplementedError):
        return True
    message = f"{type(exc).__name__}: {exc}".lower()
    mentions_stream = "stream" in message or "streaming" in message
    unsupported_markers = (
        "unsupported",
        "not support",
        "not supported",
        "not implemented",
        "cannot stream",
    )
    return mentions_stream and any(marker in message for marker in unsupported_markers)


def chunk_tool_call_deltas(chunk: Any) -> tuple[list[Any], bool]:
    choices = stream_value(chunk, "choices", []) or []
    if not choices:
        return [], False
    choice = choices[0]
    delta = stream_value(choice, "delta")
    if delta is not None:
        tool_calls = stream_value(delta, "tool_calls")
        if isinstance(tool_calls, list) and tool_calls:
            return tool_calls, False
    message = stream_value(choice, "message")
    if message is None:
        return [], False
    tool_calls = stream_value(message, "tool_calls", []) or []
    if isinstance(tool_calls, list):
        return tool_calls, True
    return [], False


def streamed_tool_calls(states: list[dict[str, Any]]) -> list[ToolCall]:
    parsed: list[ToolCall] = []
    for state in states:
        error = cast(str | None, state.get("error"))
        name_raw = state.get("name")
        name = name_raw.strip() if isinstance(name_raw, str) else ""
        if not name:
            error = append_tool_error(error, "Missing tool name")

        arguments: dict[str, Any] = {}
        arguments_obj = state.get("arguments_obj")
        if isinstance(arguments_obj, dict):
            arguments = dict(arguments_obj)
        else:
            argument_parts = state.get("arguments_parts")
            argument_text = "".join(argument_parts) if isinstance(argument_parts, list) else ""
            if argument_text:
                try:
                    parsed_arguments = json.loads(argument_text)
                    if isinstance(parsed_arguments, dict):
                        arguments = parsed_arguments
                    else:
                        error = append_tool_error(
                            error, "Tool arguments JSON must decode to object"
                        )
                except json.JSONDecodeError as exc:
                    error = append_tool_error(
                        error,
                        f"Invalid JSON in tool arguments: {exc}. Raw: {argument_text[:100]}",
                    )

        call_id_raw = state.get("id")
        call_id = call_id_raw if isinstance(call_id_raw, str) and call_id_raw else new_id()
        parsed.append(ToolCall(id=call_id, name=name, arguments=arguments, error=error))
    return parsed



from dataclasses import dataclass

from blackgeorge.core.report import Report
from blackgeorge.store.state import RunState


@dataclass(frozen=True)
class StepResult:
    report: Report
    state: RunState | None


type StepOutput = Report | StepResult
